import json
import os
from typing import Any, Dict, List, Optional, Union
from uuid import uuid4

import backoff
import requests
from nucliadb_models.metadata import ResourceProcessingStatus
from nucliadb_models.resource import Resource
from nucliadb_models.search import (
    AskRequest,
    Filter,
    RagImagesStrategies,
    RagStrategies,
)
from nucliadb_sdk.v2 import exceptions
from pydantic import BaseModel, Field, ValidationError

from nuclia import get_list_parameter, get_regional_url
from nuclia.data import get_async_auth, get_auth
from nuclia.decorators import kb, pretty
from nuclia.exceptions import RateLimitError
from nuclia.lib.kb import AsyncNucliaDBClient, NucliaDBClient
from nuclia.sdk.logger import logger

RESOURCE_ATTRIBUTES = [
    "icon",
    "origin",
    "extra",
    "conversations",
    "links",
    "texts",
    "files",
    "usermetadata",
    "fieldmetadata",
    "title",
    "summary",
    "metadata",
    "security",
    "wait_for_commit",
]


class RagStrategiesParse(BaseModel):
    rag_strategies: list[RagStrategies] = Field(default=[])


class RagImagesStrategiesParse(BaseModel):
    rag_images_strategies: list[RagImagesStrategies] = Field(default=[])


class NucliaResource:
    """
    Manage existing resource

    All commands accept either `rid` or `slug` to identify the targeted resource.
    """

    @backoff.on_exception(
        backoff.expo,
        RateLimitError,
        jitter=backoff.random_jitter,
        max_tries=5,
        factor=10,
    )
    @kb
    def create(*args, **kwargs) -> str:
        ndb: NucliaDBClient = kwargs["ndb"]
        slug = kwargs.get("slug") or uuid4().hex
        kw = {
            "kbid": ndb.kbid,
            "slug": slug,
        }
        for param in RESOURCE_ATTRIBUTES:
            if param in kwargs:
                kw[param] = kwargs.get(param)
        try:
            resource = ndb.ndb.create_resource(**kw)
        except exceptions.RateLimitError as exc:
            logger.debug(
                "Rate limited while trying to create a resource. Waiting a bit before trying again..."
            )
            raise RateLimitError() from exc
        rid = resource.uuid
        return rid

    @kb
    @pretty
    def ask(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        query: Union[str, dict, AskRequest, None] = None,
        answer_json_schema: Optional[Dict[str, Any]] = None,
        answer_json_file: Optional[str] = None,
        filters: Optional[Union[List[str], List[Filter]]] = None,
        rag_strategies: Optional[list[RagStrategies]] = None,
        rag_images_strategies: Optional[list[RagImagesStrategies]] = None,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]

        if query is None:
            query = ""

        if answer_json_file is not None:
            if os.path.exists(answer_json_file):
                with open(answer_json_file, "r") as json_file_handler:
                    answer_json_schema = json.load(json_file_handler)

        if isinstance(query, str):
            req = AskRequest(
                query=query,
                answer_json_schema=answer_json_schema,
            )
            if filters is not None:
                req.filters = filters
            if rag_strategies is not None:
                req.rag_strategies = RagStrategiesParse.model_validate(
                    {"rag_strategies": rag_strategies}
                ).rag_strategies
            if rag_images_strategies is not None:
                req.rag_images_strategies = RagImagesStrategiesParse.model_validate(
                    {"rag_images_strategies": rag_images_strategies}
                ).rag_images_strategies
        elif isinstance(query, dict):
            try:
                req = AskRequest.model_validate(query)
            except ValidationError:
                raise ValueError("Invalid AskRequest object.")

        elif isinstance(query, AskRequest):
            req = query
        else:
            raise ValueError("Invalid query type. Must be str, dict or AskRequest.")

        if rid:
            res = ndb.ndb.ask_on_resource(
                req,
                kbid=ndb.kbid,
                rid=rid,
            )
        elif slug:
            res = ndb.ndb.ask_on_resource_by_slug(
                req,
                kbid=ndb.kbid,
                slug=slug,
            )
        else:
            raise ValueError("Either rid or slug must be provided")

        return res

    @kb
    def exists(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ) -> bool:
        ndb: NucliaDBClient = kwargs["ndb"]
        url = _resource_head_url(kbid=ndb.kbid, rid=rid, slug=slug)
        resp = ndb.ndb.session.head(url)
        if resp.status_code == 404:
            return False
        resp.raise_for_status()
        return True

    @kb
    @pretty
    def get(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        show: Optional[List[str]] = None,
        extracted: Optional[List[str]] = None,
        **kwargs,
    ) -> Resource:
        show = show or ["basic"]
        extracted = extracted or []
        ndb: NucliaDBClient = kwargs["ndb"]
        show = get_list_parameter(show)
        extracted = get_list_parameter(extracted)
        if "basic" not in show:
            show.append("basic")
        if rid:
            res = ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid,
                rid=rid,
                query_params={"show": show, "extracted": extracted},
            )
        elif slug:
            res = ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid,
                slug=slug,
                query_params={"show": show, "extracted": extracted},
            )
        else:
            raise ValueError("Either rid or slug must be provided")

        if (
            "extracted" in show
            and res.metadata.status != ResourceProcessingStatus.PROCESSED
        ):
            logger.warning(
                "Resource is not processed yet, extracted content may be empty or incomplete."
            )

        return res

    @kb
    def download_file(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        file_id: str,
        output: str,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        if rid:
            res = ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid, rid=rid, query_params={"show": ["values"]}
            )
        elif slug:
            res = ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid, slug=slug, query_params={"show": ["values"]}
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        file_field = res.data.files.get(file_id)
        if not file_field:
            raise ValueError(f"File with id {file_id} not found in resource")
        url = get_regional_url(ndb.region, "/api/v1" + file_field.value.file.uri)
        download = requests.get(url, stream=True, headers=ndb.headers)
        if download.status_code != 200:
            raise ValueError(f"Error downloading file: {download.text}")
        with open(output, "wb") as f:
            for chunk in download.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

    @kb
    def temporal_download_url(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        file_id: str,
        ttl: int = 10,
        **kwargs,
    ):
        ndb = kwargs["ndb"]
        if rid:
            res = ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid, rid=rid, query_params={"show": ["values"]}
            )
        elif slug:
            res = ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid, slug=slug, query_params={"show": ["values"]}
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        file_field = res.data.files.get(file_id)
        if not file_field:
            raise ValueError(f"File with id {file_id} not found in resource")
        url = get_regional_url(ndb.region, "/api/v1" + file_field.value.file.uri)
        auth = get_auth()
        token = auth.create_ephemeral_token(ndb.kbid, ttl=ttl)
        return f"{url}?eph-token={token.token}"

    @kb
    def update(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        kw = {
            "kbid": ndb.kbid,
        }
        for param in RESOURCE_ATTRIBUTES:
            if param in kwargs:
                kw[param] = kwargs.get(param)  # type: ignore
        if rid:
            kw["rid"] = rid
            if slug:
                kw["slug"] = slug
            ndb.ndb.update_resource(**kw)
        elif slug:
            kw["rslug"] = slug
            ndb.ndb.update_resource_by_slug(**kw)
        else:
            raise ValueError("Either rid or slug must be provided")

    @kb
    def send_to_process(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        kw = {
            "kbid": ndb.kbid,
        }
        if rid:
            kw["rid"] = rid
            if slug:
                kw["slug"] = slug
            ndb.ndb.reprocess_resource(**kw)
        elif slug:
            kw["rslug"] = slug
            ndb.ndb.reprocess_resource_by_slug(**kw)
        else:
            raise ValueError("Either rid or slug must be provided")

    @kb
    def delete(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        if rid:
            ndb.ndb.delete_resource(kbid=ndb.kbid, rid=rid)
        elif slug:
            ndb.ndb.delete_resource_by_slug(kbid=ndb.kbid, rslug=slug)
        else:
            raise ValueError("Either rid or slug must be provided")


class AsyncNucliaResource:
    """
    Manage existing resource

    All commands accept either `rid` or `slug` to identify the targeted resource.
    """

    @kb
    async def create(*args, **kwargs) -> str:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        slug = kwargs.get("slug") or uuid4().hex
        kw = {
            "kbid": ndb.kbid,
            "slug": slug,
        }
        for param in RESOURCE_ATTRIBUTES:
            if param in kwargs:
                kw[param] = kwargs.get(param)
        resource = await ndb.ndb.create_resource(**kw)
        rid = resource.uuid
        return rid

    @kb
    async def exists(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ) -> bool:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        url = _resource_head_url(kbid=ndb.kbid, rid=rid, slug=slug)
        resp = await ndb.ndb.session.head(url)
        if resp.status_code == 404:
            return False
        resp.raise_for_status()
        return True

    @kb
    @pretty
    async def get(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        show: Optional[List[str]] = None,
        extracted: Optional[List[str]] = None,
        **kwargs,
    ) -> Resource:
        show = show or ["basic"]
        extracted = extracted or []
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        show = get_list_parameter(show)
        extracted = get_list_parameter(extracted)
        if "basic" not in show:
            show.append("basic")
        if rid:
            res = await ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid,
                rid=rid,
                query_params={"show": show, "extracted": extracted},
            )
        elif slug:
            res = await ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid,
                slug=slug,
                query_params={"show": show, "extracted": extracted},
            )
        else:
            raise ValueError("Either rid or slug must be provided")

        if (
            "extracted" in show
            and res.metadata.status != ResourceProcessingStatus.PROCESSED
        ):
            logger.warning(
                "Resource is not processed yet, extracted content may be empty or incomplete."
            )

        return res

    @kb
    async def download_file(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        file_id: str,
        output: str,
        **kwargs,
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if rid:
            res = await ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid, rid=rid, query_params={"show": ["values"]}
            )
        elif slug:
            res = await ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid, slug=slug, query_params={"show": ["values"]}
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        file_field = res.data.files.get(file_id)
        if not file_field:
            raise ValueError(f"File with id {file_id} not found in resource")
        url = get_regional_url(ndb.region, "/api/v1" + file_field.value.file.uri)
        download = requests.get(url, stream=True, headers=ndb.headers)
        if download.status_code != 200:
            raise ValueError(f"Error downloading file: {download.text}")
        with open(output, "wb") as f:
            for chunk in download.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

    @kb
    async def temporal_download_url(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        file_id: str,
        ttl: int = 10,
        **kwargs,
    ):
        ndb = kwargs["ndb"]
        if rid:
            res = await ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid, rid=rid, query_params={"show": ["values"]}
            )
        elif slug:
            res = await ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid, slug=slug, query_params={"show": ["values"]}
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        file_field = res.data.files.get(file_id)
        if not file_field:
            raise ValueError(f"File with id {file_id} not found in resource")
        url = get_regional_url(ndb.region, "/api/v1" + file_field.value.file.uri)
        download = requests.get(url, stream=True, headers=ndb.headers)
        if download.status_code != 200:
            raise ValueError(f"Error downloading file: {download.text}")
        auth = get_async_auth()
        token = await auth.create_ephemeral_token(ndb.kbid, ttl=ttl)
        return f"{url}?eph-token={token.token}"

    @kb
    async def send_to_process(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        kw = {
            "kbid": ndb.kbid,
        }
        if rid:
            kw["rid"] = rid
            if slug:
                kw["slug"] = slug
            await ndb.ndb.reprocess_resource(**kw)
        elif slug:
            kw["rslug"] = slug
            await ndb.ndb.reprocess_resource_by_slug(**kw)
        else:
            raise ValueError("Either rid or slug must be provided")

    @kb
    async def update(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        kw = {
            "kbid": ndb.kbid,
        }
        for param in RESOURCE_ATTRIBUTES:
            if param in kwargs:
                kw[param] = kwargs.get(param)  # type: ignore
        if rid:
            kw["rid"] = rid
            if slug:
                kw["slug"] = slug
            await ndb.ndb.update_resource(**kw)
        elif slug:
            kw["rslug"] = slug
            await ndb.ndb.update_resource_by_slug(**kw)
        else:
            raise ValueError("Either rid or slug must be provided")

    @kb
    async def delete(
        self, *, rid: Optional[str] = None, slug: Optional[str] = None, **kwargs
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if rid:
            await ndb.ndb.delete_resource(kbid=ndb.kbid, rid=rid)
        elif slug:
            await ndb.ndb.delete_resource_by_slug(kbid=ndb.kbid, rslug=slug)
        else:
            raise ValueError("Either rid or slug must be provided")


def _resource_head_url(
    *, kbid: str, rid: Optional[str] = None, slug: Optional[str] = None
) -> str:
    if rid:
        url = "v1/kb/{kbid}/resource/{rid}".format(kbid=kbid, rid=rid)
    elif slug:
        url = "v1/kb/{kbid}/slug/{slug}".format(kbid=kbid, slug=slug)
    else:
        raise ValueError("Either rid or slug must be provided")
    return url



"""Tests for the Nuclia Agent CLI interact() method."""

from typing import Type, Union
from unittest.mock import patch
from uuid import uuid4

import pytest
from rich.console import Console

from nuclia.sdk.agent import AsyncNucliaAgent, NucliaAgent
from nuclia.sdk.agent_cli import NucliaAgentCLI
from nuclia.tests.utils import maybe_await


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_simple_question(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact with a simple question and exit."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    # Simulate user asking a question and then exiting
    with patch.object(
        console, "input", side_effect=["What is Eric known for?", "/exit"]
    ):
        await maybe_await(cli.interact())

    # Check that output was generated
    output = console.export_text()
    assert len(output) > 0
    assert "Nuclia Agent CLI" in output
    assert "What is Eric known for?" in output or "Eric" in output
    assert "Goodbye" in output or "exit" in output.lower()


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_help_command(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact with help command."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=["/help", "/exit"]):
        await maybe_await(cli.interact())

    output = console.export_text()
    assert "Available Commands" in output
    assert "Exit the interactive CLI" in output
    assert "Show this help message" in output
    assert "new_session" in output or "NEW_SESSION" in output
    assert "list_sessions" in output or "LIST_SESSIONS" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_empty_input(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact handles empty input gracefully."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=["", "  ", "/exit"]):
        await maybe_await(cli.interact())

    # Should not crash and should show welcome
    output = console.export_text()
    assert "Nuclia Agent CLI" in output
    assert "Goodbye" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_unknown_command(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact handles unknown commands."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=["/unknown_command", "/exit"]):
        await maybe_await(cli.interact())

    output = console.export_text()
    assert "Unknown command" in output
    assert "/unknown_command" in output or "unknown_command" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_clear_command(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact with clear command."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=["/clear", "/exit"]):
        await maybe_await(cli.interact())

    # Should clear and show welcome again
    output = console.export_text()
    assert "Nuclia Agent CLI" in output
    assert "Goodbye" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_list_sessions(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact with list sessions command."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=["/list_sessions", "/exit"]):
        await maybe_await(cli.interact())

    output = console.export_text()
    assert "Available Sessions" in output
    assert "Goodbye" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_keyboard_interrupt(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact handles keyboard interrupt gracefully."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(console, "input", side_effect=KeyboardInterrupt()):
        await maybe_await(cli.interact())

    output = console.export_text()
    # Should handle interrupt gracefully
    assert "Nuclia Agent CLI" in output
    assert "Interrupted" in output or "Goodbye" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_multiple_questions(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test CLI interact with multiple questions in sequence."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    with patch.object(
        console,
        "input",
        side_effect=[
            "What is Eric known for?",
            "Tell me about Python",
            "/exit",
        ],
    ):
        await maybe_await(cli.interact())

    output = console.export_text()
    assert "Nuclia Agent CLI" in output
    # Should have processed both questions
    assert "Eric" in output or "What is Eric" in output
    assert "Goodbye" in output


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_cli_interact_session_workflow(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    """Test complete session workflow: create, list, use, switch."""
    console = Console(record=True, force_terminal=False, width=120)
    cli = NucliaAgentCLI(console=console)

    session_name = f"nuclia-py-cli-test-{uuid4().hex}"
    with patch.object(
        console,
        "input",
        side_effect=[
            "/new_session",
            session_name,
            "What is Eric known for?",
            "/list_sessions",
            "/change_session",
            "ephemeral",
            "What does he enjoy?",
            "/delete_session",
            session_name,
            "/exit",
        ],
    ):
        await maybe_await(cli.interact())

    output = console.export_text()
    assert "Nuclia Agent CLI" in output
    # Should show session creation
    assert session_name in output or "created" in output.lower()
    # Should process both questions
    assert "humor" in output
    # Should show session list
    assert "list" in output.lower() or "sessions" in output.lower()



import inspect
from typing import AsyncIterator, Awaitable, Iterator, TypeVar, Union

T = TypeVar("T")


async def maybe_await(aw_or_value: Union[Awaitable[T], T]) -> T:
    """Given a result from a call to a sync or an async function (a value or an
    awaitable of a value), await if needed and return the value. This is useful
    to test sync and async versions of the SDK using the same test functions.

    """
    if inspect.isawaitable(aw_or_value):
        return await aw_or_value
    else:
        return aw_or_value


async def maybe_async_iterate(
    iterator: Union[AsyncIterator[T], Iterator[T]],
) -> AsyncIterator[T]:
    """Given an async iterator or a regular iterator, yield items in an async manner.
    This is useful to test sync and async versions of the SDK using the same test functions.

    """
    if isinstance(iterator, AsyncIterator):
        # It's an async iterator
        async for item in iterator:
            yield item
    else:
        # It's a regular iterator
        for item in iterator:
            yield item



import base64
import io
import tempfile
import warnings
from time import sleep
from typing import Type, Union

import httpx
import pytest
from nucliadb_models.common import File
from nucliadb_models.file import FileField
from nucliadb_sdk.v2.exceptions import NotFoundError

from nuclia.sdk.resource import AsyncNucliaResource, NucliaResource
from nuclia.tests.utils import maybe_await


def test_resource(testing_config):
    nresource = NucliaResource()
    try:
        res = nresource.get(slug="res1")
        nresource.delete(rid=res.id)
    except NotFoundError:
        pass

    assert not nresource.exists(slug="res1")

    res_id = nresource.create(slug="res1")

    assert nresource.exists(slug="res1")
    assert nresource.exists(rid=res_id)

    res = nresource.get(rid=res_id)
    assert res

    res = nresource.get(slug="res1")
    assert res
    assert res.id == res_id

    assert res.title == "res1"

    nresource.update(
        rid=res_id,
        title="My great resource",
        texts={"text1": {"body": "Hello here"}},
    )
    res = nresource.get(slug="res1", show=["basic", "values"])

    assert res
    assert res.title in ["My great resource", "res1"]
    if res.data.texts:
        assert res.data.texts["text1"]
    else:
        warnings.warn("No texts found in resource")

    nresource.delete(rid=res_id)

    try:
        sleep(0.5)
        nresource.get(slug="res1")
        assert False
    except NotFoundError:
        assert True


def test_resource_download(testing_config):
    nresource = NucliaResource()
    slug = "download-file-test"

    # Clean up any existing resource with the same slug
    try:
        res = nresource.get(slug=slug)
    except NotFoundError:
        pass
    else:
        nresource.delete(rid=res.id)

    # Create a resource with a file
    file_binary = io.BytesIO(b"some random data")
    res_id = nresource.create(
        slug=slug,
        files={
            "file": FileField(
                file=File(
                    filename="image.png",
                    content_type="image/png",
                    payload=base64.b64encode(file_binary.getvalue()).decode("utf-8"),
                )
            )
        },
    )

    # Now download the file
    with tempfile.TemporaryDirectory() as tmpdir:
        output = f"{tmpdir}/image.png"
        nresource.download_file(rid=res_id, file_id="file", output=output)


def test_resource_get_public_url(testing_config):
    nresource = NucliaResource()
    slug = "download-file-test"

    # Clean up any existing resource with the same slug
    try:
        res = nresource.get(slug=slug)
    except NotFoundError:
        pass
    else:
        nresource.delete(rid=res.id)

    # Create a resource with a file
    content = b"some random data"
    file_binary = io.BytesIO(content)
    res_id = nresource.create(
        slug=slug,
        files={
            "file": FileField(
                file=File(
                    filename="image.png",
                    content_type="image/png",
                    payload=base64.b64encode(file_binary.getvalue()).decode("utf-8"),
                )
            )
        },
    )

    url = nresource.temporal_download_url(rid=res_id, file_id="file", ttl=60)
    assert "eph-token=" in url
    downloaded = httpx.get(url)
    assert downloaded.status_code == 200
    assert downloaded.content == content


@pytest.mark.asyncio
async def test_resource_get_public_url_async(testing_config):
    from nuclia.sdk.resource import AsyncNucliaResource

    nresource = AsyncNucliaResource()
    slug = "download-file-test-async"

    # Clean up any existing resource with the same slug
    try:
        res = await nresource.get(slug=slug)
    except NotFoundError:
        pass
    else:
        await nresource.delete(rid=res.id)

    # Create a resource with a file
    content = b"some random data"
    file_binary = io.BytesIO(content)
    res_id = await nresource.create(
        slug=slug,
        files={
            "file": FileField(
                file=File(
                    filename="image.png",
                    content_type="image/png",
                    payload=base64.b64encode(file_binary.getvalue()).decode("utf-8"),
                )
            )
        },
    )

    url = await nresource.temporal_download_url(rid=res_id, file_id="file", ttl=60)
    assert "eph-token=" in url
    async with httpx.AsyncClient() as client:
        downloaded = await client.get(url)
    assert downloaded.status_code == 200
    assert downloaded.content == content


@pytest.mark.parametrize(
    "nresource_klass",
    [NucliaResource, AsyncNucliaResource],
)
async def test_resource_crud_by_id(
    testing_config,
    nresource_klass: Union[Type[NucliaResource], Type[AsyncNucliaResource]],
):
    nresource = nresource_klass()
    slug = "crud-by-id"

    try:
        await maybe_await(nresource.delete(slug=slug))
    except NotFoundError:
        pass

    assert not await maybe_await(nresource.exists(slug=slug))

    rid = await maybe_await(
        nresource.create(
            slug=slug,
            title="Testing",
            texts={"mytext": {"body": "testing is essential for a reliable software"}},
        )
    )

    assert await maybe_await(nresource.exists(slug=slug))
    assert await maybe_await(nresource.exists(rid=rid))

    res = await maybe_await(nresource.get(rid=rid))
    assert res.title == "Testing"

    await maybe_await(nresource.update(rid=rid, title="Reliability"))
    res = await maybe_await(nresource.get(rid=rid))
    assert res.title == "Reliability"

    await maybe_await(nresource.delete(rid=rid))
    try:
        for _ in range(3):
            await maybe_await(nresource.get(slug=slug))
            sleep(0.5)
        assert False
    except NotFoundError:
        assert True


@pytest.mark.parametrize(
    "nresource_klass",
    [NucliaResource, AsyncNucliaResource],
)
async def test_resource_crud_by_slug(
    testing_config,
    nresource_klass: Union[Type[NucliaResource], Type[AsyncNucliaResource]],
):
    nresource = nresource_klass()
    slug = "crud-by-slug"

    try:
        await maybe_await(nresource.delete(slug=slug))
    except NotFoundError:
        pass

    await maybe_await(
        nresource.create(
            slug=slug,
            title="Testing",
            texts={"mytext": {"body": "testing is essential for a reliable software"}},
        )
    )
    res = await maybe_await(nresource.get(slug=slug))
    assert res.title == "Testing"

    await maybe_await(nresource.update(slug=slug, title="Reliability"))
    res = await maybe_await(nresource.get(slug=slug))
    assert res.title == "Reliability"

    await maybe_await(nresource.delete(slug=slug))
    try:
        for _ in range(3):
            await maybe_await(nresource.get(slug=slug))
            sleep(0.5)
        assert False
    except NotFoundError:
        assert True


@pytest.mark.parametrize(
    "nresource_klass",
    [NucliaResource, AsyncNucliaResource],
)
async def test_resource_file_download(
    testing_config,
    nresource_klass: Union[Type[NucliaResource], Type[AsyncNucliaResource]],
):
    nresource = nresource_klass()
    slug = "resource-with-file"

    try:
        await maybe_await(nresource.delete(slug=slug))
    except NotFoundError:
        pass

    rid = await maybe_await(
        nresource.create(
            slug=slug,
            files={
                "myfile": {
                    "language": "en",
                    "file": {
                        "filename": "myfile.txt",
                        "payload": base64.b64encode(b"this is my file"),
                    },
                }
            },
        )
    )

    res = await maybe_await(nresource.get(rid=rid))
    assert res

    with tempfile.NamedTemporaryFile() as fp:
        await maybe_await(
            nresource.download_file(rid=rid, file_id="myfile", output=fp.name)
        )

        with open(fp.name, "r") as f:
            content = f.read()
            assert content == "this is my file"

    # cleanup test

    await maybe_await(nresource.delete(rid=rid))
    try:
        for _ in range(3):
            await maybe_await(nresource.get(slug=slug))
            sleep(0.5)
        assert False
    except NotFoundError:
        assert True



class ConfigNotFound(Exception):
    pass


class NeedUserToken(Exception):
    pass


class UserTokenExpired(Exception):
    pass


class NotDefinedDefault(Exception):
    pass


class APIException(Exception):
    api_name: str

    def __init__(self, code: int, detail: str):
        self.code = code
        self.detail = detail
        message = f"Exception calling {self.api_name} API: {self.code} {self.detail}"
        super().__init__(message)


class RaoAPIException(APIException):
    api_name: str = "RAO"


class NuaAPIException(APIException):
    api_name: str = "NUA"


class AlreadyConsumed(Exception):
    pass


class GettingRemoteFileError(Exception):
    pass


class RateLimitError(Exception):
    pass


class NucliaConnectionError(Exception):
    pass


class KBNotAvailable(Exception):
    pass


class DuplicateError(Exception):
    pass


class InvalidPayload(Exception):
    pass
